@@ -10,12 +10,8 @@
 """
 
 import inspect
-try:
-    from urlparse import urlsplit
-except ImportError:
-    from urllib.parse import urlsplit
 
-from flask import request, current_app
+from flask import request, current_app, flash
 from flask_wtf import Form as BaseForm
 from wtforms import TextField, PasswordField, validators, \
     SubmitField, HiddenField, BooleanField, ValidationError, Field
@@ -23,7 +19,7 @@ from flask_login import current_user
 from werkzeug.local import LocalProxy
 
 from .confirmable import requires_confirmation
-from .utils import verify_and_update_password, get_message, config_value
+from .utils import verify_and_update_password, get_message, config_value, validate_redirect_url
 
 # Convenient reference
 _datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)
@@ -137,12 +133,10 @@ class NextFormMixin():
     next = HiddenField()
 
     def validate_next(self, field):
-        if field.data:
-            url_next = urlsplit(field.data)
-            url_base = urlsplit(request.host_url)
-            if url_next.netloc and url_next.netloc != url_base.netloc:
-                field.data = ''
-                raise ValidationError(get_message('INVALID_REDIRECT')[0])
+        if field.data and not validate_redirect_url(field.data):
+            field.data = ''
+            flash(*get_message('INVALID_REDIRECT'))
+            raise ValidationError(get_message('INVALID_REDIRECT')[0])
 
 
 class RegisterFormMixin():
@@ -209,6 +203,8 @@ class LoginForm(Form, NextFormMixin):
 
     def __init__(self, *args, **kwargs):
         super(LoginForm, self).__init__(*args, **kwargs)
+        if not self.next.data:
+            self.next.data = request.args.get('next', '')
         self.remember.default = config_value('DEFAULT_REMEMBER_ME')
 
     def validate(self):
@@ -223,7 +219,6 @@ class LoginForm(Form, NextFormMixin):
             self.password.errors.append(get_message('PASSWORD_NOT_PROVIDED')[0])
             return False
 
-
         self.user = _datastore.get_user(self.email.data)
 
         if self.user is None:
@@ -276,9 +271,6 @@ class ChangePasswordForm(Form, PasswordFormMixin):
         if not super(ChangePasswordForm, self).validate():
             return False
 
-        if self.password.data.strip() == '':
-            self.password.errors.append(get_message('PASSWORD_NOT_PROVIDED')[0])
-            return False
         if not verify_and_update_password(self.password.data, current_user):
             self.password.errors.append(get_message('INVALID_PASSWORD')[0])
             return False
